#include "ootcp.h"

pthread_t input_ootcp_thread;
pthread_t output_ootcp_thread;
void * input_ootcp_thread_code(void * t);
void * output_ootcp_thread_code(void * t);
int input_fifo_handler(unsigned int);
int output_fifo_handler(unsigned int);

pthread_mutex_t mutex;	// used when acessing Connections index
pthread_mutex_t mutexIndex[MAX_CONS*2];	// used by all connections
int inThreadFifo, outThreadFifo;


int init_module(void) {

	int i;

	rtl_printf("OOTCP Loading...\n");

	// Initialize rtfifos for thread communication
	// Create fifo for input thread
	for(i = 0 ; i < RTF_NO ; i++){ // find the first idle fifo...
		if(rtf_create(i, 50000) == OK){
			inThreadFifo = i; // ...and reserve it
			break;
		}
	}

	if(i == RTF_NO){
		printf("OOTCP: Error in creating rt-fifo for input thread\n");
		return 0;
	}
	else
		register_ootcp_fifo(inThreadFifo);

	// Create fifo for output thread
	for(i = inThreadFifo + 1 ; i < RTF_NO ; i++){ // find the first idle fifo...
		if(rtf_create(i, 50000) == OK){
			outThreadFifo = i; // ...and reserve it
			break;
		}
	}

	if(i == RTF_NO){
		printf("OOTCP: Error in creating rt-fifo for output thread\n");
		return 0;
	}

	// Create handlers for the thread's fifos
	rtf_create_rt_handler(inThreadFifo, input_fifo_handler);
	rtf_create_rt_handler(outThreadFifo, output_fifo_handler);

	// Initialize all mutexes
	pthread_mutex_init(&mutex, NULL);

	for(i = 0 ; i < (MAX_CONS*2) ; i++)
		pthread_mutex_init((mutexIndex + i), NULL);

	// Initialize OOTCP's threads
	pthread_create(&input_ootcp_thread, NULL, input_ootcp_thread_code, 0);
	pthread_create(&output_ootcp_thread, NULL, output_ootcp_thread_code, 0);
	
	rtl_printf("OOTCP Loaded.\n");
	return 0;
}


int cleanup_module(void) {

	int i;

	pthread_delete_np(input_ootcp_thread);
	pthread_delete_np(output_ootcp_thread);

	register_ootcp_fifo(-1);
	rtf_destroy(inThreadFifo);
	rtf_destroy(outThreadFifo);

	pthread_mutex_destroy(&mutex);

	for(i = 0 ; i < (MAX_CONS*2) ; i++)
		pthread_mutex_destroy(mutexIndex + i);

	rtl_printf("OOTCP unloaded.\n");

	return 0;
} 


// Thread for handling incoming segments
void * input_ootcp_thread_code(void * t) {

	InputSegment * inSeg;
	Connection * con;
	RSTSegment rstSeg;

	struct sched_param p;
	p.sched_priority = 5;
	pthread_setschedparam(pthread_self(), SCHED_RR, &p);

	while(1){
		if(rtf_get(inThreadFifo, &inSeg, sizeof(InputSegment *)) == 0){ // wait for incoming segment
			pthread_suspend_np(pthread_self());
			rtf_get(inThreadFifo, &inSeg, sizeof(InputSegment *));
		}

		if(inSeg->tcpHead.flags == SYN)
			con = ConnectionHandler::findListeningConnection(inSeg->ipHead.targetIP, inSeg->tcpHead.targetPort);
		else
			con = ConnectionHandler::findConnection(inSeg->ipHead.targetIP, inSeg->tcpHead.targetPort, inSeg->ipHead.sourceIP, inSeg->tcpHead.sourcePort);

		if(con != NULL)		
			con->takeSegment(inSeg);
		else{
			printf("OOTCP: No connection matching incoming segment\n");
			rstSeg.setHeaders(inSeg);
			transmit_packet((TCPKT *)&rstSeg);
			release_packet((TCPKT *)inSeg);
		}
	}

	return (void *) NULL;
}


// Thread for handling outgoing segments
void * output_ootcp_thread_code(void * t) {

	Connection *con, *newCon;
	InputSegment *inSeg;
	OutputSegment outSeg;
	RSTSegment rstSeg;
	int dataLength, tries, prstTries, timeout;
	char prst = 'p';

	struct sched_param p;
	p.sched_priority = 6;
	pthread_setschedparam(pthread_self(), SCHED_RR, &p);

	while(1){
		if(rtf_get(outThreadFifo, &con, sizeof(Connection *)) == 0){ // wait for incoming segment
			pthread_suspend_np(pthread_self());
			rtf_get(outThreadFifo, &con, sizeof(Connection *));
		}

		switch(con->status){

		case ESTABLISHED:
		case SYN_RCVD:
		case CLOSE_WAIT:

			switch(con->nextStatus){

				case FIN_WAIT_1:
				case LAST_ACK:
					outSeg.setHeaders(con, FIN+ACK, 0);
					con->status = con->nextStatus;
					con->outSequence++;
					transmit_packet((TCPKT *)&outSeg);
					con->timer.start(con->timeout);

					if(con->status == CLOSED){
						ConnectionHandler::deallocate(con);
						break;
					}
					else
						if(con->status == FIN_WAIT_2)
							break;
					
					tries = 0;
					while((tries < 5) && ((con->status != CLOSED) || (con->status != FIN_WAIT_2)))
						tries += con->retransmit(&outSeg, 5-tries);

					if(con->status != FIN_WAIT_2){
						con->status = CLOSED;
						ConnectionHandler::deallocate(con);
					}
					break;

				case DATA_SENT:	// send data

					timeout = 3000;
					tries = prstTries = 0;

					while(((con->outBuf.in != con->outBuf.out) || con->outBuf.full) && (tries < 6) && (con->status != CLOSED)){

						while((con->outWindow == 0) && (prstTries < 5) && (con->status != CLOSED)){
							if(con->timer.start(timeout) == TIMEDOUT){
								outSeg.data = &prst;
								outSeg.setHeaders(con, ACK, 1); // create "persist" segment
								transmit_packet((TCPKT *)&outSeg); // send "persist" segment
								timeout *= 2;
								prstTries++;
							}
							else
								prstTries = 0;
						}

						while(((con->outBuf.trial != con->outBuf.in) || con->outBuf.newData) && (con->outWindow > 0) && (con->status != CLOSED)){
							dataLength = con->outMaxSegSize < con->outWindow ? con->outMaxSegSize : con->outWindow;
							pthread_mutex_lock(con->outMutex);
							dataLength = con->outBuf.get(&outSeg.data, dataLength);
							outSeg.setHeaders(con, ACK, dataLength);
							con->outSequence += dataLength;
							con->outWindow -= dataLength;
							pthread_mutex_unlock(con->outMutex);
							transmit_packet((TCPKT *)&outSeg);
						}

						while(((con->outBuf.trial != con->outBuf.out) || (con->outBuf.full && !con->outBuf.newData)) && (con->status != CLOSED)){
							if(con->timer.start(timeout) == TIMEDOUT){
								pthread_mutex_lock(con->outMutex);
								if(con->outBuf.trial > con->outBuf.out){
									con->outSequence -= (con->outBuf.trial - con->outBuf.out);
									con->outWindow += (con->outBuf.trial - con->outBuf.out);
								}
								else{
									con->outSequence -= (MAX_BUF_SIZE - (LWORD)con->outBuf.out + (LWORD)con->outBuf.trial);
									con->outWindow += (MAX_BUF_SIZE - (WORD)con->outBuf.out + (WORD)con->outBuf.trial);
								}
								con->outBuf.trial = con->outBuf.out;
								pthread_mutex_unlock(con->outMutex);
								timeout *= 2;
								tries++;
								break;
							}
							else
								tries = 0;
						}

					}
					
					if(tries == 6){
						con->status = CLOSED;
						ConnectionHandler::deallocate(con);
					}
		
					break;

				case ACK_SENT:
					if(con->timer.start(500) == TIMEDOUT){ // wait for 500msec before sending ACK
						outSeg.setHeaders(con, ACK, 0);
						transmit_packet((TCPKT *)&outSeg);
					}
					break;

				default:
					break;
			}
			break;

		case LISTEN:
			pthread_suspend_np(pthread_self());
			if(rtf_get(outThreadFifo, &inSeg, sizeof(InputSegment *)) == 0)
				break;
			if((newCon = ConnectionHandler::allocate(con->appThread)) == NULL){
				rstSeg.setHeaders(inSeg);
				transmit_packet((TCPKT *)&rstSeg);
				release_packet((TCPKT *)inSeg);
				break;
			}
			newCon->outMaxSegSize = inSeg->getMaxSegSize();
			newCon->inSequence = inSeg->tcpHead.sequence + 1;
			newCon->nextStatus = SYN_RCVD;
			if(newCon->activeOpen(con->localIP, con->localPort, inSeg->ipHead.sourceIP, inSeg->tcpHead.sourcePort) == OK){
				pthread_mutex_lock(con->inMutex);
				con->inBuf.append((char *)&newCon, sizeof(Connection *));
				con->currNumOfReqs++;
				pthread_mutex_unlock(con->inMutex);
				pthread_wakeup_np(con->appThread);
			}
			else
				ConnectionHandler::deallocate(newCon);
			release_packet((TCPKT *)inSeg);
			break;
				
		default:
			break;
		}
	}

	return (void *) NULL;

}


int Connection::activeOpen(char * localIP, int localPort, char * remoteIP, int remotePort){

	if((localPort < 0) || (localPort > 65536) || (remotePort < 0) || (remotePort > 65536))
		return ERROR;

	return activeOpen(string2LwordIP(localIP), (WORD)localPort, string2LwordIP(remoteIP), (WORD)remotePort);
}


int Connection::activeOpen(LWORD localIP, WORD localPort, LWORD remoteIP, WORD remotePort){

	if(status != CLOSED)
		return ERROR;

	pthread_mutex_lock(&mutex);	// ensures that the connection will be unique
	if(ConnectionHandler::findConnection(localIP, localPort, remoteIP, remotePort) != NULL){
		pthread_mutex_unlock(&mutex);
		return ERROR;	// duplicate connection already exists
	}
	this->localIP = localIP;
	this->localPort = localPort;
	this->remoteIP = remoteIP;
	this->remotePort = remotePort;
	pthread_mutex_unlock(&mutex);

	outSequence = (LWORD)this;	// random 32bit number
	inWindow = MAX_BUF_SIZE;
	timeout = 3000; // milliseconds
	ackPending = false;
	zeroInWin = false;
	// initialize input buffer
	inBuf.in = inBuf.out = inBuf.buffer;
	inBuf.full = false;
	// initialize output buffer
	outBuf.in = outBuf.out = outBuf.trial = outBuf.buffer;
	outBuf.full = outBuf.newData = false;
	outBuf.lastAck = 0;

	OutputSegment outSeg;

	switch(nextStatus){
	case SYN_RCVD:
		outSeg.setHeaders(this, SYN+ACK, 0);
		outSequence++;
		transmit_packet((TCPKT *)&outSeg);
		status = SYN_RCVD;
		break;
	default:
		inSequence = 0;
		outSeg.setHeaders(this, SYN, 0);
		outSequence++;
		status = SYN_SENT;
		transmit_packet((TCPKT *)&outSeg);
		break;
	}

	timer.start(timeout);

	if(status == ESTABLISHED)
		return OK;

	if(status == CLOSED) // connection was remotely rejected
		return ERROR;

	int retries = 0;
	while((retries < 2) && (status != ESTABLISHED) && (status != CLOSED))
			retries += retransmit(&outSeg, 2 - retries);
	
	if(status != ESTABLISHED){
		status = CLOSED;
		return ERROR;
	}
	else
		return OK;
}


int Connection::passiveOpen(char * localIP, int localPort, int numOfRequests){

	if((localPort < 0) || (localPort > 65536))
		return ERROR;

	return passiveOpen(string2LwordIP(localIP), (WORD)localPort, numOfRequests);
}


int Connection::passiveOpen(LWORD localIP, WORD localPort, int numOfRequests){

	if(status != CLOSED)
		return ERROR;

	pthread_mutex_lock(&mutex); // ensures that the connection will be unique
	if(ConnectionHandler::findListeningConnection(localIP, localPort) != NULL){
		pthread_mutex_unlock(&mutex);
		return ERROR;	// duplicate connection already exists
	}
	this->localIP = localIP;
	this->localPort = localPort;
	pthread_mutex_unlock(&mutex);

	inBuf.in = inBuf.out = inBuf.buffer;
	inBuf.full = false;

	maxNumOfReqs = numOfRequests;
	currNumOfReqs = 0;
	status = LISTEN;

	return OK;
}


Connection * Connection::accept(void){

	if(status != LISTEN)
		return NULL;

	Connection * newCon;

	pthread_mutex_lock(inMutex);
	while(inBuf.get((char *)&newCon, sizeof(Connection *)) == 0){
		pthread_mutex_unlock(inMutex);
		pthread_suspend_np(pthread_self());
		pthread_mutex_lock(inMutex);
	}
	currNumOfReqs--;
	pthread_mutex_unlock(inMutex);

	return newCon;
}


int Connection::send(char * message, int dataLength){

	if(dataLength <= 0)
		return ERROR;

	if(status != ESTABLISHED)
		return ERROR;

	int outDataLength;
	pthread_mutex_lock(outMutex);
	while((outDataLength = outBuf.append(message, dataLength)) == 0){
		pthread_mutex_unlock(outMutex);
		pthread_suspend_np(pthread_self());
		pthread_mutex_lock(outMutex);
	}
	pthread_mutex_unlock(outMutex);

	if(nextStatus == ACK_SENT)
		timer.stop();

	Connection * con = this;
	if(outBuf.trial == outBuf.out){ // if not already sending...
		nextStatus = DATA_SENT;
		rtf_put(outThreadFifo, &con, sizeof(Connection *));
	}

	return outDataLength;

}


int Connection::receive(char * buffer, int dataLength){

	if(dataLength <= 0)
		return ERROR;

	if((status == CLOSE_WAIT) || (status == CLOSED))
		return 0;

	int inDataLength;

	pthread_mutex_lock(inMutex);
	while((inDataLength = inBuf.get(buffer, dataLength)) == 0){
		pthread_mutex_unlock(inMutex);
		pthread_suspend_np(pthread_self());
		if((status == CLOSE_WAIT) || (status == CLOSED))
			return 0;
		pthread_mutex_lock(inMutex);
	}
	inWindow += inDataLength;
	pthread_mutex_unlock(inMutex);
	
	OutputSegment ackSeg;
	if(zeroInWin){
		if(inWindow > (MAX_SEG_SIZE - 40 < MAX_BUF_SIZE / 2 ? MAX_SEG_SIZE - 40 : MAX_BUF_SIZE / 2)){ 
			ackSeg.setHeaders(this, ACK, 0);
			transmit_packet((TCPKT *)&ackSeg); // advertise window
			zeroInWin = false;
		}
	}

	return inDataLength;
}


int Connection::close(void){

	Connection * con = this;

	switch(status){

	case LISTEN:
	case SYN_SENT:
		status = CLOSED;
		ConnectionHandler::deallocate(this);
		break;

	case CLOSED:
		ConnectionHandler::deallocate(this);
		break;

	case ESTABLISHED:
	case SYN_RCVD:
		nextStatus = FIN_WAIT_1;
		rtf_put(outThreadFifo, &con, sizeof(Connection *));
		break;

	case CLOSE_WAIT:
		nextStatus = LAST_ACK;
		rtf_put(outThreadFifo, &con, sizeof(Connection *));
		break;

	default:
		return ERROR;
	}

	return OK;
}


int Connection::retransmit(OutputSegment * outSeg, int retries){

	int i = 0, ret = TIMEDOUT, incr_timeout = timeout;

	while((i < retries) && (ret == TIMEDOUT)){
		transmit_packet((TCPKT *)outSeg);
		incr_timeout *= 2;
		ret = timer.start(incr_timeout);
		i++;
	}

	return i;
}



void Connection::takeSegment(InputSegment * inSeg){

	Connection * thisCon = this;
	RSTSegment rstSeg;
	OutputSegment ackSeg;

	if(status == LISTEN){
		if(maxNumOfReqs == currNumOfReqs){
			rstSeg.setHeaders(inSeg);
			transmit_packet((TCPKT *)&rstSeg);
			release_packet((TCPKT *)inSeg);
			return;
		}
		rtf_put(outThreadFifo, &thisCon, sizeof(Connection *));
		rtf_put(outThreadFifo, &inSeg, sizeof(InputSegment *));
		release_packet((TCPKT *)inSeg);
		return;
	}

	if(inSequence == 0)
		inSequence = inSeg->tcpHead.sequence;

	if(inSeg->tcpHead.sequence > inSequence){	// if "future" segment...
		ackSeg.setHeaders(this, ACK, 0); // inform sender of the correct sequence
		transmit_packet((TCPKT *)&ackSeg);
		release_packet((TCPKT *)inSeg);
		return;
	}

	if(inSeg->tcpHead.sequence == inSequence){	// if correct sequence...

		inSeg->tcpHead.flags &= (FIN + SYN + RST + ACK);

		if(inSeg->tcpHead.flags & RST){ // force close
			status = CLOSED;
			timer.stop();
			release_packet((TCPKT *)inSeg);
			pthread_wakeup_np(appThread);
			return;
		}

		switch(status){

		case SYN_SENT:
			if((inSeg->tcpHead.flags == SYN + ACK) && (inSeg->tcpHead.ack == outSequence)){
				outMaxSegSize = inSeg->getMaxSegSize();
				outWindow = inSeg->tcpHead.windowSize;
				inSequence++;
				ackSeg.setHeaders(this, ACK, 0);
				transmit_packet((TCPKT *)&ackSeg);
				status = ESTABLISHED;
				timer.stop();
			}
			else{
				inSequence = 0;
				rstSeg.setHeaders(inSeg);
				transmit_packet((TCPKT *)&rstSeg); // send RST
				release_packet((TCPKT *)inSeg);
				return;
			}
			break;

		case SYN_RCVD:
			if((inSeg->tcpHead.flags == ACK) && (inSeg->tcpHead.ack == outSequence)){
				outWindow = inSeg->tcpHead.windowSize;
				status = ESTABLISHED;
				timer.stop();
			}
			else{
				inSequence = 0;
				rstSeg.setHeaders(inSeg);
				transmit_packet((TCPKT *)&rstSeg); // send RST
				release_packet((TCPKT *)inSeg);
				return;
			}
			break;

		case ESTABLISHED:
			if(inSeg->tcpHead.flags & ACK){
				pthread_mutex_lock(outMutex);
				outWindow = inSeg->tcpHead.windowSize;
				if(outBuf.ack(outSequence, inSeg->tcpHead.ack) == OK)
					pthread_wakeup_np(appThread);
				pthread_mutex_unlock(outMutex); 
				timer.stop(); // stop possible waiting for ACK
			}
			if((inSeg->tcpHead.flags & FIN) && (inSeg->tcpHead.ack == outSequence)){
				inSequence++;
				ackSeg.setHeaders(this, ACK, 0);
				transmit_packet((TCPKT *)&ackSeg);
				status = CLOSE_WAIT;
				pthread_wakeup_np(appThread);
			}
			break;

		case CLOSE_WAIT:
			if(inSeg->tcpHead.flags & ACK){
				pthread_mutex_lock(outMutex);
				outWindow = inSeg->tcpHead.windowSize;
				if(outBuf.ack(outSequence, inSeg->tcpHead.ack) == OK)
					pthread_wakeup_np(appThread);
				pthread_mutex_unlock(outMutex);
				timer.stop(); // stop possible waiting for ACK
			}
			break;

		case LAST_ACK:
			if((inSeg->tcpHead.flags == ACK) && (inSeg->tcpHead.ack == outSequence)){
				status = CLOSED;
				timer.stop();
			}
			break;

		case FIN_WAIT_1:
			if(inSeg->tcpHead.flags & FIN){
				inSequence++;
				if((inSeg->tcpHead.flags & ACK) && (inSeg->tcpHead.ack == outSequence)){
					status = CLOSED; // should be timed-wait
					timer.stop();
				}
				else if(!(inSeg->tcpHead.flags & ACK) || (inSeg->tcpHead.ack < outSequence))
					status = CLOSING;
				ackSeg.setHeaders(this, ACK, 0);
				transmit_packet((TCPKT *)&ackSeg);
			}
			else if((inSeg->tcpHead.flags & ACK) && (inSeg->tcpHead.ack == outSequence)){
				status = FIN_WAIT_2;
				timer.stop();
			}
			break;

		case CLOSING:
			if((inSeg->tcpHead.flags & ACK) && (inSeg->tcpHead.ack == outSequence)){
					status = CLOSED; // should be timed-wait
					timer.stop();
			}
			break;

		case FIN_WAIT_2:
			if(inSeg->tcpHead.flags & FIN){
				inSequence++;
				ackSeg.setHeaders(this, ACK, 0);
				transmit_packet((TCPKT *)&ackSeg);
				status = CLOSED; // should be timed-wait
				ConnectionHandler::deallocate(this);
			}
			break;

		default:
			break;
		}
	}

	// Calculate dataLength and compare with inWindow
	int dataLength = inSeg->ipHead.segmentLength - (inSeg->tcpHead.headerLength >> 2) - 20;
	char * data = (char *)&inSeg->data;

	if((dataLength > 0) && (inSeg->tcpHead.sequence < inSequence)){	// if "past" segment...
		dataLength -= (inSequence - inSeg->tcpHead.sequence); // read new data (if any)
		if(dataLength > 0)
			data += (inSequence - inSeg->tcpHead.sequence);
	}

	dataLength = dataLength < inWindow ? dataLength : inWindow;

	// Process data (if any)
	if(dataLength > 0){
		pthread_mutex_lock(inMutex);
		inBuf.append(data, dataLength);
		inWindow -= dataLength;
		if(inWindow == 0)
			zeroInWin = true;
		inSequence += dataLength;
		pthread_mutex_unlock(inMutex);
		pthread_wakeup_np(appThread);
		if(outBuf.trial == outBuf.out){ // if not sending...
			if(!ackPending){
				nextStatus = ACK_SENT;		// wait for 500msec before sending ACK
				rtf_put(outThreadFifo, &thisCon, sizeof(Connection *));
				ackPending = true;
			}
			else{ 
				timer.stop(); // stop pending ACK
				ackSeg.setHeaders(this, ACK, 0);
				transmit_packet((TCPKT *)&ackSeg);
				ackPending = false;
			}
		}
	}

	release_packet((TCPKT *)inSeg);

}


char * Connection::getLocalIP(void){

	return lword2StringIP(localIP);
}

char * Connection::getRemoteIP(void){

	return lword2StringIP(remoteIP);
}

int Connection::getLocalPort(void){

	return (int)localPort;
}

int Connection::getRemotePort(void){

	return (int)remotePort;
}

char * Connection::lword2StringIP(LWORD IPAddr){

	static char stringIP[16];
	BYTE byte[4] = { 0 };
	int i, j, k;

	strcpy(stringIP, "000.000.000.000");

	byte[0] = (BYTE)((IPAddr & 0xff000000L) >> 24);
	byte[1] = (BYTE)((IPAddr & 0x00ff0000L) >> 16);
	byte[2] = (BYTE)((IPAddr & 0x0000ff00L) >> 8);
	byte[3] = (BYTE)(IPAddr & 0x000000ffL);

	for(i = 0 ; i < 4 ; i++){
		k = i * 4;
		j = 100;
		while(byte[i] != 0){
			while(byte[i] / j >= 1){
				stringIP[k]++;
				byte[i] -= j;
			}
			k++;
			j /= 10;
		}
	}
	
	return stringIP;
}

LWORD Connection::string2LwordIP(char * IPAddr){
	
	BYTE byte[4] = { 0 };
	char * ch = IPAddr;
	int i, j = 1, k = 3;

	while(*ch != '\0')
		ch++;
	ch--;
	while(ch >= IPAddr){
		switch (*ch) {
                case '0':           i = 0;  break;
                case '1':           i = 1;  break;
                case '2':           i = 2;  break;
                case '3':           i = 3;  break;
                case '4':           i = 4;  break;
                case '5':           i = 5;  break;
                case '6':           i = 6;  break;
                case '7':           i = 7;  break;
                case '8':           i = 8;  break;
                case '9':           i = 9;  break;
                case 'A': case 'a': i = 10; break;
                case 'B': case 'b': i = 11; break;
                case 'C': case 'c': i = 12; break;
                case 'D': case 'd': i = 13; break;
                case 'E': case 'e': i = 14; break;
                case 'F': case 'f': i = 15; break;
                default:                    break;
            }
		byte[k] += (i * j);
		ch--;
		j *= 10;
		if(*ch == '.'){
			k--;
			ch--;
			j = 1;
		}
	}

    return((byte[0]<<24) | (byte[1]<<16) | (byte[2]<<8) | byte[3]);
}


void OutputSegment::setHeaders(Connection * con, BYTE flags, int dataLength){

	// Set IP Header
	ipHead.sourceIP = con->localIP;
	ipHead.targetIP = con->remoteIP;
	ipHead.segmentLength = dataLength + 40;  // will change if we have options
	// the rest of the IP Header will be initialized by IP itself

	// Set TCP Header
	tcpHead.sourcePort = con->localPort;
	tcpHead.targetPort = con->remotePort;
	tcpHead.sequence = con->outSequence;
	tcpHead.ack = con->inSequence;
	tcpHead.flags = flags;
	tcpHead.headerLength = 20 << 2;  // will change if we have options
	tcpHead.windowSize = con->inWindow;
	tcpHead.checksum = 0;  // checksum is calculated later
	tcpHead.urgentPointer = 0;

	// send max seg size
	Option * op;
	if((flags & SYN) && (dataLength == 0)){
		ipHead.segmentLength = 44;
		tcpHead.headerLength = 24 << 2;
		op = (Option *)&data;
		op->kind = 2;
		op->length = 4;
		op->data = swapw(MAX_SEG_SIZE - 40);
	}
}


void RSTSegment::setHeaders(InputSegment * inSeg){

	// Initialize IP Header
	ipHead.sourceIP = inSeg->ipHead.targetIP;
	ipHead.targetIP = inSeg->ipHead.sourceIP;
	ipHead.segmentLength = 40;
	// the rest of the IP Header will be initialized by IP itself

	// Initialize TCP Header
	tcpHead.sourcePort = inSeg->tcpHead.targetPort;
	tcpHead.targetPort = inSeg->tcpHead.sourcePort;
	tcpHead.sequence = inSeg->tcpHead.ack;

	int inDataLength = inSeg->ipHead.segmentLength - (inSeg->tcpHead.headerLength >> 2) - 20;
	tcpHead.ack = inSeg->tcpHead.sequence + inDataLength;
	if(inSeg->tcpHead.flags & (SYN + FIN))
		tcpHead.ack++;

	tcpHead.flags = RST + ACK;
	tcpHead.headerLength = 20 << 2;
	tcpHead.windowSize = 0;
	tcpHead.checksum = 0;  // checksum is calculated later
	tcpHead.urgentPointer = 0;
}


WORD InputSegment::getMaxSegSize(void){

	Option * op = (Option *)&data;

	if(((tcpHead.headerLength >> 2) >= 24) && (op->kind == 2) && (op->length == 4))
		return swapw(op->data);
	else
		return MAX_SEG_SIZE - 40;
}


int Timer::start(int millisecs){

	int i = 0;

	blocked = true;
	while((i++ < millisecs) && blocked)
		usleep(1000);	// sleep for a millisecond

	if(blocked)
		return TIMEDOUT;
	else
		return OK;
}


void Timer::stop(void){

	blocked = false;
}



Connection * ConnectionHandler::findConnection(LWORD localIP, WORD localPort, LWORD remoteIP, WORD remotePort){
	
	int i = 0;

	while(i < MAX_CONS){
		if((conIndex[i].reserved) && (conIndex[i].status != CLOSED)){
			if((conIndex[i].localIP == localIP) && (conIndex[i].localPort == localPort) && (conIndex[i].remoteIP == remoteIP) && (conIndex[i].remotePort == remotePort))
				return &conIndex[i];
		}
		i++;
	}
	return NULL;  // Connection not found
}


Connection * ConnectionHandler::findListeningConnection(LWORD localIP, WORD localPort){
	
	int i = 0;

	while(i < MAX_CONS){
		if((conIndex[i].reserved) && (conIndex[i].status == LISTEN)){
			if((conIndex[i].localIP == localIP) && (conIndex[i].localPort == localPort))
				return &conIndex[i];
		}
		i++;
	}
	return NULL;  // Connection not found
}


Connection * ConnectionHandler::allocate(pthread_t appThread){

	int i = 0;

	pthread_mutex_lock(&mutex);
	while(conIndex[i].reserved && (i < MAX_CONS))
		i++;

	if(i == MAX_CONS){	// all connections reserved
		pthread_mutex_unlock(&mutex);
		return NULL;
	}

	conIndex[i].reserved = true;
	pthread_mutex_unlock(&mutex);

	conIndex[i].appThread = appThread;
	conIndex[i].inMutex = mutexIndex + 2*i;
	conIndex[i].outMutex = mutexIndex + 2*i + 1;

	return &conIndex[i];

}

int ConnectionHandler::deallocate(Connection * con){

	OutputSegment rstSeg;
	int i = 0;

	while((&conIndex[i] != con) && (i < MAX_CONS))
		i++;

	if(i == MAX_CONS)
		return ERROR;	// connection not found

	if(con->status != CLOSED){ // force shutdown
		rstSeg.setHeaders(con, RST+ACK, 0);
		transmit_packet((TCPKT *)&rstSeg);
		con->status = CLOSED;
	}

	conIndex[i].reserved = false;

	return OK;
}


void ConnectionHandler::printStatus(void){

	printf("Local Address            Foreign Address            Status\n");

	int i = 0;
	char status[15], localIP[16], remoteIP[16];

	while(i < MAX_CONS){
		if((conIndex[i].reserved) && (conIndex[i].status != LISTEN)){
			switch(conIndex[i].status){
			case SYN_SENT:
				strcpy(status, "SYN SENT");
				break;
			case SYN_RCVD:
				strcpy(status, "SYN RCVD");
				break;
			case ESTABLISHED:
				strcpy(status, "ESTABLISHED");
				break;
			case CLOSE_WAIT:
				strcpy(status, "CLOSE WAIT");
				break;
			case LAST_ACK:
				strcpy(status, "LAST ACK");
				break;
			case FIN_WAIT_1:
				strcpy(status, "FIN WAIT 1");
				break;
			case FIN_WAIT_2:
				strcpy(status, "FIN WAIT 2");
				break;
			case CLOSING:
				strcpy(status, "CLOSING");
				break;
			case TIMED_WAIT:
				strcpy(status, "TIMED WAIT");
				break;
			default:
				break;
			}
			strcpy(localIP, conIndex[i].lword2StringIP(conIndex[i].localIP));
			strcpy(remoteIP, conIndex[i].lword2StringIP(conIndex[i].remoteIP));
			printf("%s:%d       %s:%d       %s\n", localIP, conIndex[i].localPort, remoteIP, conIndex[i].remotePort, status);
		}
		i++;
	}

}


int input_fifo_handler(unsigned int){
	pthread_wakeup_np(input_ootcp_thread);
	return 0;
}

int output_fifo_handler(unsigned int){
	pthread_wakeup_np(output_ootcp_thread);
	return 0;
}


int CircularBuffer::append(char * source, int dataLength){

	if(full)
		return 0;

	newData = true;

	dataLength = dataLength < MAX_BUF_SIZE ? dataLength : MAX_BUF_SIZE;

	if((in >= out) && (dataLength <= buffer + MAX_BUF_SIZE - in)){
		memcpy(in, source, dataLength);
		in += dataLength;
		if(in == buffer + MAX_BUF_SIZE)
			in = buffer;
		if(in == out)
			full = true;
		return dataLength;
	}

	int tempDataLength = dataLength;

	if(in >= out){
		memcpy(in, source, buffer + MAX_BUF_SIZE - in);
		tempDataLength -= (buffer + MAX_BUF_SIZE - in);
		source += (buffer + MAX_BUF_SIZE - in);
		in = buffer;
	}

	if(out - in > tempDataLength){
		memcpy(in, source, tempDataLength);
		in += tempDataLength;
		tempDataLength = 0;
	}
	else{
		memcpy(in, source, out - in);
		tempDataLength -= (out - in);
		in = out;
		full = true;
	}

	return (dataLength - tempDataLength); // return buffered data
}


int OutputBuffer::get(char ** target, int dataLength){

	if(in > trial)
		dataLength = (in - trial) < dataLength ? (in - trial) : dataLength;
	else
		dataLength = (buffer + MAX_BUF_SIZE - trial) < dataLength ? (buffer + MAX_BUF_SIZE - trial) : dataLength;

	*target = trial;
	trial += dataLength;

	if(trial == buffer + MAX_BUF_SIZE)
		trial = buffer;

	newData = false;

	return dataLength;
}


int InputBuffer::get(char * target, int dataLength){

	if((in == out) && (!full))
		return 0;	// buffer is empty

	full = false;

	if(in > out){
		dataLength = (in - out) < dataLength ? (in - out) : dataLength;
		memcpy(target, out, dataLength);
		out += dataLength;
		return dataLength;
	}

	int tempDataLength = dataLength;
	tempDataLength -= (buffer + MAX_BUF_SIZE - out) < dataLength ? (buffer + MAX_BUF_SIZE - out) : dataLength;
	memcpy(target, out, dataLength - tempDataLength);
	out += (dataLength - tempDataLength);
	if(out == buffer + MAX_BUF_SIZE)
		out = buffer;
	if(tempDataLength)
		tempDataLength -= get(target + dataLength - tempDataLength, tempDataLength);

	return (dataLength - tempDataLength);
}


int OutputBuffer::ack(LWORD outSequence, LWORD ack){
		
	if((ack > outSequence) || (ack <= lastAck))
		return ERROR;

	int notAcked = outSequence - ack;

	if(notAcked == 0)
		out = trial; // everything sent is acked
	else{
		if(trial - buffer >= notAcked)
			out = trial - notAcked;
		else{
			notAcked -= (trial - buffer);
			out = buffer + MAX_BUF_SIZE - notAcked;
		}
	}
	full = false;
	lastAck = ack;

	return OK;
}

		





