#include <rtl_cpp.h>
#include <rtl_fifo.h>
#include <string.h>
#include "tcpkt.h"

extern "C" {
#include <rtl.h>
}

#define printf rtl_printf

typedef enum { CLOSED, LISTEN, SYN_SENT, SYN_RCVD, ESTABLISHED, CLOSE_WAIT, LAST_ACK, FIN_WAIT_1, FIN_WAIT_2, CLOSING, TIMED_WAIT, DATA_SENT, ACK_SENT } CON_STATUS;

#define ERROR		-1
#define TIMEDOUT	-1
#define OK			0

#define FIN        0x01    /* Option flags: no more data */
#define SYN        0x02    /*           sync sequence nums */
#define RST        0x04    /*           reset connection */
#define PSH        0x08    /*           push buffered data */
#define ACK        0x10    /*           acknowledgement */
#define URG        0x20    /*           urgent */

#define MAX_CONS	20		// maximum connections

#define MAX_SEG_SIZE  1500		// maximum Ethernet segment size
#define MAX_BUF_SIZE  32000     // maximum output buffer size;


//****************************************************
// forward declarations:

class OutputSegment;
class InputSegment;
class RSTSegment;
class ConnectionHandler;

//****************************************************

class Timer{

	bool blocked;

public:
	int start(int millisecs);
	void stop(void);
};

class CircularBuffer{

	friend class Connection;
	friend void * output_ootcp_thread_code(void * t);

protected:
	char * in;
	char * out;
	char buffer[MAX_BUF_SIZE];
	bool full;
	bool newData;

public:
	int append(char * source, int dataLength);
};


class InputBuffer: public CircularBuffer{

	friend class Connection;

public:
	int get(char * target, int dataLength);
};


class OutputBuffer: public CircularBuffer{

	friend class Connection;
	friend void * output_ootcp_thread_code(void * t);

	char * trial;
	LWORD lastAck;

public:
	int get(char ** target, int dataLength);
	int ack(LWORD outSequence, LWORD ack);
};

class Connection{

	friend class OutputSegment;
	friend class ConnectionHandler;
	friend void * output_ootcp_thread_code(void * t);
	
	LWORD localIP, 
		  remoteIP;
	WORD localPort, 
	     remotePort;
	CON_STATUS status,
		   nextStatus;
	Timer timer;
	int timeout; // in seconds
	WORD outWindow;
	WORD inWindow;
	WORD outMaxSegSize;
	LWORD outSequence;
	LWORD inSequence;
	int maxNumOfReqs;
	int currNumOfReqs;
	pthread_mutex_t *inMutex, 
		            *outMutex;
	InputBuffer inBuf;
	OutputBuffer outBuf;
	bool ackPending;
	bool zeroInWin;
	bool reserved;
	pthread_t appThread;

	int retransmit(OutputSegment * outSeg, int retries);
	int activeOpen(LWORD localIP, WORD localPort, LWORD remoteIP, WORD remotePort);
	int passiveOpen(LWORD localIP, WORD localPort, int numOfRequests);
	LWORD string2LwordIP(char * IPAddr);
	char * lword2StringIP(LWORD IPAddr);

public:
	Connection(void){ status = CLOSED; reserved = false; };
	int activeOpen(char * localIP, int localPort, char * remoteIP, int remotePort);
	int passiveOpen(char * localIP, int localPort, int numOfRequests);
	Connection * accept(void);
	int send(char * message, int dataLength);
	int receive(char * buffer, int dataLength);
	char * getLocalIP(void);
	char * getRemoteIP(void);
	int getLocalPort(void);
	int getRemotePort(void);
	int close(void);
	void takeSegment(InputSegment * inSeg);

};


class ConnectionHandler{

	friend void * input_ootcp_thread_code(void * t);
	friend class Connection;

	static Connection conIndex[MAX_CONS];

	static Connection * findConnection(LWORD localIP, WORD localPort, LWORD remoteIP, WORD remotePort);
	static Connection * findListeningConnection(LWORD localIP, WORD localPort);

public:
	static Connection * allocate(pthread_t appThread);
	static int deallocate(Connection * con);
	static void printStatus(void);
};


//************************************************

class IPHeader{

public:
    BYTE  verHeadLength,
          service;
    WORD  segmentLength,
          id,               
          fragments;           
    BYTE  timeToLive,             
          protocolType;            
    WORD  headerChecksum;    
	LWORD sourceIP;
	LWORD targetIP;
};


class TCPHeader{

public:
	WORD sourcePort;
	WORD targetPort;
	LWORD sequence;
	LWORD ack;
	BYTE headerLength;
	BYTE flags;
	WORD windowSize;
	WORD checksum;
	WORD urgentPointer;
};


class TCPSegment{

public:
	IPHeader ipHead;
	TCPHeader tcpHead;
	char * data;
};


class Option{

public:
	BYTE kind;
	BYTE length;
	WORD data;
};

//************************************************

class InputSegment : public TCPSegment{ 

public:
	WORD getMaxSegSize(void);
};


class OutputSegment : public TCPSegment{

public:
	OutputSegment(void){ };
	void setHeaders(Connection * con, BYTE flags, int dataLength);
};


class RSTSegment : public OutputSegment{

public:
	RSTSegment(void){ };
	void setHeaders(InputSegment * inSeg);
};


//************************************************

extern "C"{
int transmit_packet(TCPKT * tcpkt);
int release_packet(TCPKT * tcpkt);
int register_ootcp_fifo(int fifo_id);
WORD swapw(WORD w);
}
